var USER_DATA = "user_data_363";
var PLAYED_GAME_DATA = "play_data_9999";
var AUTH_TOKEN = "user_auth_token_2163";
var SERVER_HISTORY = "server_history_9919";

angular.module('experticity.userService', ['ionic'])

/**
 * User Service
 *
 * Responsible for maintaining state about the user both on
 * the device, and on the server.
 *
 * All public methods use promises.
 *
 * Public Methods:
 *  - syncProfile
 *  - login(string username, string password)
 *  - profile
 *  - syncPlayHistory
 *  - activeQuizzes
 *  - futureQuizzes
 *  - pastQuizzes
 */
.service('UserService', function ($q, $http, $cordovaLocalNotification, QuizService) {
  /**
   * Getters and setters for the main data.
   */
  var getUser = function () {
    return getObject(USER_DATA);
  };

  var setUser = function (user) {
    setObject(USER_DATA, user);
  };

  var getAuthToken = function () {
    return getObject(AUTH_TOKEN);
  };

  var setAuthToken = function (token) {
    setObject(AUTH_TOKEN, token);
  };

  var getAuthHeaders = function () {
    var token = getAuthToken();
    if (typeof token !== "string" || (token.length < 5)) {
      return {
        headers: {},
      };
    } else {
      return {
        headers: {
          "Authorization": "JWT " + token,
        }
      };
    }
  };

  /**
   * Private Methods
   */
  var getPlayedGames = function () {
    var data = getObject(PLAYED_GAME_DATA);
    if (Object.keys(data).length === 0) { data = []; }

    var played_games = [];

    for (var x = 0; x < data.length; x++) {
      var played_obj = data[x];
      var pg = PlayedGame.prototype.createPlayedGame(played_obj.quiz, played_obj.play_time, played_obj.answers);

      pg.completed = played_obj.completed;
      played_games.push(pg);
    }

    return played_games;
  };

  var addPlayedGame = function (playedGame) {
    var games = getPlayedGames();
    games.push(playedGame);
    setObject(PLAYED_GAME_DATA, games);
  };

  // Used in the activeQuizzes to check if a number is /not/ in an array.
  var checkNumberNotInArray = function(arr, number) {
    if (arr.indexOf(number) == -1) {
      return number;
    }
    else {
      return 0;
    }
  };

  var under18 = function () {
    var user = getUser();

    if ((user.contestant || {}).birthdate === undefined) {
      return false;
    }

    var age = moment(user.contestant.birthdate);
    var now = moment();
    return now.diff(age, 'years') < 18;
  };
  /**
   * End Private Methods
   */

  return {
    /**
     * Get the latest data from the server, and store it
     * on the device. Also, refresh their auth token.
     *
     * @return {object} A dictionary containing
     *                    - {'result': 'synced'}
     */
    syncProfile: function () {
      console.log("UserService.sync");
      var dfd = $q.defer();

      var user_url = SERVER_ADDRESS + "user/";
      $http.get(user_url, getAuthHeaders()).then(
        function (response) {
          setUser(response.data);
          dfd.resolve({'result': 'synced'});
        },
        function (error) {
          dfd.reject(error);
        }
      );

      // We need to refresh the auth token.
      var refresh_url = SERVER_ADDRESS + "auth/refresh/";
      var token = {
        token: getAuthToken(),
      };

      $http.post(refresh_url, token, getAuthHeaders()).then(
        function (response) {
          setAuthToken(response.data.token);
        }
      );

      return dfd.promise;
    },
    /**
     * Get a auth token for the username specified.
     * Also, ensure that the profile has been synced to the device.
     *
     * @param  {string} username
     * @param  {string} password
     *
     * @return {object} A dictionary containing
     *                    - {'result': {'errors': 'Invalid Password'}}
     *                    - {'result': 'Successful'}
     */
    login: function (username, password) {
      console.log("UserService.login(" + username + "," + password + ")");
      var dfd = $q.defer();

      var auth_url = SERVER_ADDRESS + "auth/login/";
      data = { 'username': username, 'password': password };

      var ctx = this;
      $http.post(auth_url, data).
        success(function (data, status, headers, config) {
          setAuthToken(data.token);

          ctx.profile().then(
            function (profile) {
              dfd.resolve(profile);
            }
          );
        }).
        error(function (data, status, headers, config) {
          dfd.reject({'result':data});
        });

      return dfd.promise;
    },
    logout: function () {
      this.syncPlayHistory().then().finally(
        function () {
          setObject(USER_DATA, {});
          setObject(AUTH_TOKEN, "");
          setObject(PLAYED_GAME_DATA, []); // The users profile.
          setObject(SAVED_PLAYED_GAMES, []); // The global object.
          setObject(EVENT_DATA_SERVICE, []);
          QuizService.sync();
        }
      );
    },
    /**
     * Used to register with the server.
     *
     * @param  {object} data the attributes of this new user.
     *
     * @return {object} an object describing the register attempt.
     *                    - {'result': {'errors': 'Invalid Username'}}
     *                    - {'result': 'Successful'}
     */
    register: function (data) {
      var dfd = $q.defer();

      var registration_url = SERVER_ADDRESS + "user/register";

      console.log(data);
      $http.post(registration_url, data).
        success(function (data, status, headers, config) {
          setUser(data);
          setAuthToken(data.token);

          dfd.resolve({'result': 'successful'});
        }).
        error(function (data, status, headers, config) {
          dfd.reject({'result': data});
        });

      return dfd.promise;
    },
    /**
     * Used to get and set a profiles data.
     * `changed_data` is an optional param used
     * to updated any new data. If changed_data is present,
     * the data will be synced to the server.
     *
     * @param  {object} changed_data an object of any changed data.
     *
     * @return {object} object of the users profile.
     *                         - {
     *                         			'profile': [object]
     *                         			'errors': '[username]'
     *                         		}
     */
    profile: function (changed_data) {
      var dfd = $q.defer();

      if (changed_data !== undefined) {
        var current_user = getUser();
        var update_url = SERVER_ADDRESS + "user/" + current_user.id;

        changed_data.id = current_user.id;
        $http.patch(update_url, changed_data, getAuthHeaders()).
          success(function (data, status, headers, config) {
            setUser(data);
            dfd.resolve({
              'profile':getUser(),
              'errors':null,
              'under18':under18(),
            });
          }).
          error(function (data, status, headers, config) {
            dfd.reject({
              'profile':getUser(),
              'errors':data,
              'under18':under18(),
            });
          });
      } else {
        if (Object.keys(getUser()).length === 0) {
          this.syncProfile().then(
            function (data) {
              dfd.resolve({
                'profile':getUser(),
                'errors':null,
                'under18':under18(),
              });
            },
            function (error) {
              dfd.resolve({
                'profile': null,
                'under18':under18(),
                'errors': error,
              });
            }
          );
        } else {
          dfd.resolve({
            'profile':getUser(),
            'under18':under18(),
          });
        }
      }

      return dfd.promise;
    },
    /**
     * A call it and forget it function.
     * Sends the users playhistory to the server.
     *
     * @return {object} when it finsihes.
     */
    syncPlayHistory: function () {
      console.log("syncPlayHistory()");
      var dfd = $q.defer();

      var score_url = SERVER_ADDRESS + "quiz/result/";
      var played_games = PlayedGame.prototype.getSaved();

      (function(ctx, games, index) {
        if (games.length === 0) {
          dfd.resolve({'result':'finished'});
        } else {
          sendItem(games, index);
        }

        function sendItem (games, index) {
          var game = games[index];
          // If the game has been sent OR the game has not been completed.
          if (game.sent_to_server || !game.completed) {
            if (index + 1 < games.length) {
              sendItem(games, index + 1);
            } else {
              dfd.resolve({'result':'finished'});
            }
          } else {
            var d = game.exportJson();
            $http.post(score_url, d, getAuthHeaders()).
              success(function (data, status, headers, config) {
                game.sent_to_server = true;
              }).
              error(function (data, status, headers, config) {
                // Assume the connection failed.
                // But, if the status is 406 Unaccepctable.
                // And Error is set right. Set it to sent.
                game.sent_to_server = false;
                if (status === 409) {
                  if (data.error == 'already-submitted') {
                    console.log("The Server responded that this has been sent already");
                    game.sent_to_server = true;
                  }
                }
              }).finally(function () {
                setObject("7_SAVED_PLAYED_GAMES", {games: games});
                if (index + 1 < games.length) {
                  sendItem(games, index + 1);
                } else {
                  dfd.resolve({'result':'finished'});
                }
              });
          }
        }
      }) (this, played_games, 0);

      return dfd.promise;
    },
    activeQuizzes: function (event) {
      var dfd = $q.defer();

      (function (event, user) {
          var active = event.getQuizzes().active;
          var played = getPlayedGames();
          var playable_out = [];
          var played_out = [];
          var current_user = getUser();

          var synced_played_games = (((current_user||{}).contestant||{}).stats||{}).played_quizzes || [];

          for (var x = 0; x < active.length; x++) {
                var active_quiz = active[x];
                active_quiz.played = false;

                for (var y = 0; y < played.length; y++) {
                  played_quiz = played[y];

                  if (active_quiz.id == played_quiz.quiz.id) {
                    active_quiz.played = true;
                    break;
                  }
                }

                if (synced_played_games.indexOf(active_quiz.id) >= 0) {
                    active_quiz.played = true;
                }

                if (active_quiz.played) {
                    played_out.push(active_quiz);
                } else {
                    playable_out.push(active_quiz);
                }
              }

          playable_out.sort(function (a, b) {
              if (a.start_date != b.start_date) {
                  return a.start_date > b.start_date ? 1 : -1;
              } else {
                  return a.start_time > b.start_time ? 1 : -1;
              }

              return 0;
          });

          var out = playable_out.concat(played_out);

          dfd.resolve(out);
      }) (event, this);

      return dfd.promise;
    },
    futureQuizzes: function (event) {
      var dfd = $q.defer();

      (function (event, user) {
        var inactive = event.getQuizzes().future;
        var played = getPlayedGames();

        for (var x = 0; x < inactive.length; x++) {
          var inactive_quiz = inactive[x];

          inactive_quiz.played = false;
          for (var y = 0; y < played.length; y++) {
            played_quiz = played[y].quiz;

            if (inactive_quiz.id == played_quiz.id) {
              // They have played this quiz.
              inactive_quiz.played = true;
              break;
            }
          }
        }

        dfd.resolve(inactive);
      }) (event, this);

      return dfd.promise;
    },
    pastQuizzes: function (event) {
      var dfd = $q.defer();

      (function (event, user) {
        var inactive = event.getQuizzes().past;
        var played = getPlayedGames();

        for (var x = 0; x < inactive.length; x++) {
          var inactive_quiz = inactive[x];

          inactive_quiz.played = false;
          for (var y = 0; y < played.length; y++) {
            played_quiz = played[y].quiz;

            if (inactive_quiz.id == played_quiz.id) {
              // They have played this quiz.
              inactive_quiz.played = true;
              break;
            }
          }
        }

        dfd.resolve(inactive);
      }) (event, this);

      return dfd.promise;
    },
    scheduleNotifications: function (event) {
      console.log('>>>Schedule');
      var dfd = $q.defer();

      (function (event) {
        var future = event.getQuizzes().future;

        for (var x = 0; x < future.length; x++) {
          console.log(future[x].start_date + ' ' + future[x].start_time);
          moment_start = moment(future[x].start_date + ' ' + future[x].start_time);

          console.log('>>>Schedule '+future[x].name + ' at ' + moment_start );
          try {
              $cordovaLocalNotification.add({
                  id: future[x].id,
                  at: moment_start.toDate(),
                  message: "Fire up that expertise.",
                  title: "A new quiz is open.",
                  autoCancel: true
              }).then(function () {
                  console.log("The notification has been set");
              });
          } catch (err) {
              console.log("Notification Failure");
          }
        }


        dfd.resolve(future);
      }) (event);

      return dfd.promise;
    },
    getPlayHistory: function () {
      var dfd = $q.defer();

      var current_user = getUser();
      var user_games_played_url = SERVER_ADDRESS + "user/history/" + current_user.id;

      $http.get(user_games_played_url).then(
        function (data, status, headers, config) {
          dfd.resolve({"history": data.data, "type": "server"});
        },
        function (error_data, status, headers, config) {
          /**
           * Get their local play history.
           */
          var local_history = PlayedGame.prototype.getSaved();

          QuizService.getEvents().then(function (events) {
            var event_id;
            var sameEventIdFilter = function (obj) {
              return (event_id == obj.event.id);
            };

            var hasPlayedAQuizFilter = function (obj) {
              return (obj.results.length > 0);
            };

            var wonAPrizeInEventFilter = function (obj) {
              for (var x = 0; x < obj.results.length; x++) {
                if (obj.results[x].canGetPrize()) {
                  return true;
                }
              }

              return false;
            };

            var displayable_history = [];
            for (var x = 0; x < events.data.length; x++) {
              event_id = events.data[x].id;
              var event = {
                id: events.data[x].id,
                name: events.data[x].name,
                grand_prize: events.data[x].grand_prize,
                results: []
              };
              event.results = local_history.filter(sameEventIdFilter);
              displayable_history.push(event);
            }

            displayable_history = displayable_history.filter(hasPlayedAQuizFilter).filter(wonAPrizeInEventFilter);

            dfd.resolve({"history": displayable_history, "type": "local"});
          });

        }
      );

      return dfd.promise;
    },
    getAuthHeaders: function () {
      return getAuthHeaders();
    },
    addPlayedGame: function (game) {
      addPlayedGame(game);
    },
    getPlayedGames: function () {
      return getPlayedGames();
    },
    /**
     * Determine if a user can play anymore quizzes.
     * Using current logic, if they have played 1 quiz, without a
     * profile, they cannot.
     *
     * @return {promise} boolean true if they can play quizzes, false if not.
     */
    ableToPlayQuizzes: function () {
      var dfd = $q.defer();

      var able_to_play = function () {
        var system_stored_games = PlayedGame.prototype.getSaved();
        var user = getUser();

        var played_a_game = system_stored_games.length > 1;
        var has_a_profile = user.id !== null;

        if (played_a_game && has_a_profile === false) {
          dfd.resolve(false);
        } else {
          dfd.resolve(true);
        }
      };
      able_to_play();

      return dfd.promise;
    },
  };
});
